@@ -40,7 +40,7 @@ from localstack.services.awslambda.lambda_executors import (
 from localstack.services.awslambda.multivalue_transformer import multi_value_dict_for_list
 from localstack.utils.common import (
     to_str, to_bytes, load_file, save_file, TMP_FILES, ensure_readable, short_uid, json_safe,
-    mkdir, unzip, is_zip_file, zip_contains_jar_entries, run, first_char_to_lower,
+    mkdir, unzip, is_zip_file, run, first_char_to_lower,
     timestamp_millis, now_utc, safe_requests, FuncThread, isoformat_milliseconds, synchronized)
 from localstack.utils.analytics import event_publisher
 from localstack.utils.http_utils import parse_chunked_data
@@ -74,6 +74,8 @@ INVALID_PARAMETER_VALUE_EXCEPTION = 'InvalidParameterValueException'
 
 VERSION_LATEST = '$LATEST'
 
+FUNCTION_MAX_SIZE = 69905067
+
 BATCH_SIZE_RANGES = {
     'kinesis': (100, 10000),
     'dynamodb': (100, 1000),
@@ -121,6 +123,9 @@ LAMBDA_EXECUTOR = lambda_executors.AVAILABLE_EXECUTORS.get(config.LAMBDA_EXECUTO
 IAM_POLICY_VERSION = '2012-10-17'
 POLICY_NAME_PATTERN = 'lambda_policy_%s_%s'
 
+# Whether to check if the handler function exists while creating lambda function
+CHECK_HANDLER_ON_CREATION = False
+
 # Marker name to indicate that a bucket represents the local file system. This is used for testing
 # Serverless applications where we mount the Lambda code directly into the container from the host OS.
 BUCKET_MARKER_LOCAL = '__local__'
@@ -139,13 +144,15 @@ class ClientError(Exception):
 
 
 class LambdaContext(object):
-    def __init__(self, func_details, qualifier=None):
+
+    def __init__(self, func_details, qualifier=None, context=None):
         self.function_name = func_details.name()
         self.function_version = func_details.get_qualifier_version(qualifier)
-
+        self.client_context = context.get('client_context')
         self.invoked_function_arn = func_details.arn()
         if qualifier:
             self.invoked_function_arn += ':' + qualifier
+        self.cognito_identity = context.get('identity')
 
     def get_remaining_time_in_millis(self):
         # TODO implement!
@@ -269,7 +276,7 @@ def message_attributes_to_lower(message_attrs):
 
 def process_apigateway_invocation(func_arn, path, payload, stage, api_id, headers={},
                                   resource_path=None, method=None, path_params={},
-                                  query_string_params=None, request_context={}):
+                                  query_string_params=None, request_context={}, event_context={}):
     try:
         resource_path = resource_path or path
         path_params = dict(path_params)
@@ -288,13 +295,14 @@ def process_apigateway_invocation(func_arn, path, payload, stage, api_id, header
             'requestContext': request_context,
             'stageVariables': get_stage_variables(api_id, stage),
         }
-        return run_lambda(event=event, context={}, func_arn=func_arn,
+        LOG.debug('Running Lambda function %s from API Gateway invocation: %s %s' % (func_arn, method or 'GET', path))
+        return run_lambda(event=event, context=event_context, func_arn=func_arn,
             asynchronous=not config.SYNCHRONOUS_API_GATEWAY_EVENTS)
     except Exception as e:
         LOG.warning('Unable to run Lambda function on API Gateway message: %s %s' % (e, traceback.format_exc()))
 
 
-def process_sns_notification(func_arn, topic_arn, subscription_arn, message,
+def process_sns_notification(func_arn, topic_arn, subscription_arn, message, message_id,
         message_attributes, unsubscribe_url, subject='',):
     event = {
         'Records': [{
@@ -303,7 +311,7 @@ def process_sns_notification(func_arn, topic_arn, subscription_arn, message,
             'EventSubscriptionArn': subscription_arn,
             'Sns': {
                 'Type': 'Notification',
-                'MessageId': str(uuid.uuid4()),
+                'MessageId': message_id,
                 'TopicArn': topic_arn,
                 'Subject': subject,
                 'Message': message,
@@ -339,6 +347,11 @@ def process_kinesis_records(records, stream_name):
                         {
                             'eventID': 'shardId-000000000000:{0}'.format(rec['sequenceNumber']),
                             'eventSourceARN': stream_arn,
+                            'eventSource': 'aws:kinesis',
+                            'eventVersion': '1.0',
+                            'eventName': 'aws:kinesis:record',
+                            'invokeIdentityArn': 'arn:aws:iam::{0}:role/lambda-role'.format(TEST_AWS_ACCOUNT_ID),
+                            'awsRegion': aws_stack.get_region(),
                             'kinesis': rec
                         }
                         for rec in chunk
@@ -522,10 +535,12 @@ def run_lambda(event, context, func_arn, version=None, suppress_output=False, as
         func_details = ARN_TO_LAMBDA.get(func_arn)
         if not func_details:
             return not_found_error(msg='The resource specified in the request does not exist.')
-        if not context:
-            context = LambdaContext(func_details, version)
+
+        context = LambdaContext(func_details, version, context)
+
         result = LAMBDA_EXECUTOR.execute(func_arn, func_details, event, context=context,
             version=version, asynchronous=asynchronous, callback=callback)
+
     except Exception as e:
         exc_type, exc_value, exc_traceback = sys.exc_info()
         response = {
@@ -617,7 +632,7 @@ def get_handler_function_from_name(handler_name, runtime=LAMBDA_DEFAULT_RUNTIME)
 
 def error_response(msg, code=500, error_type='InternalFailure'):
     LOG.warning(msg)
-    return aws_responses.flask_error_response(msg, code=code, error_type=error_type)
+    return aws_responses.flask_error_response_json(msg, code=code, error_type=error_type)
 
 
 def get_zip_bytes(function_code):
@@ -627,6 +642,7 @@ def get_zip_bytes(function_code):
     :param function_code: https://docs.aws.amazon.com/lambda/latest/dg/API_FunctionCode.html
     :returns: bytes of the Zip file.
     """
+    function_code = function_code or {}
     if 'S3Bucket' in function_code:
         s3_client = aws_stack.connect_to_service('s3')
         bytes_io = BytesIO()
@@ -639,7 +655,7 @@ def get_zip_bytes(function_code):
         zip_file_content = function_code['ZipFile']
         zip_file_content = base64.b64decode(zip_file_content)
     else:
-        raise ClientError('No valid Lambda archive specified.')
+        raise ClientError('No valid Lambda archive specified: %s' % list(function_code.keys()))
     return zip_file_content
 
 
@@ -746,9 +762,8 @@ def set_function_code(code, lambda_name, lambda_cwd=None):
         if not is_zip_file(zip_file_content):
             raise ClientError(
                 'Uploaded Lambda code for runtime ({}) is not in Zip format'.format(runtime))
-        # Unzipping should only be required for (1) non-Java Lambdas, or (2) zip files containing JAR files
-        if not is_java or zip_contains_jar_entries(zip_file_content, 'lib/'):
-            unzip(tmp_file, lambda_cwd)
+        # Unzip the Lambda archive contents
+        unzip(tmp_file, lambda_cwd)
 
     # Obtain handler details for any non-Java Lambda function
     if not is_java:
@@ -756,7 +771,8 @@ def set_function_code(code, lambda_name, lambda_cwd=None):
         handler_function = get_handler_function_from_name(handler_name, runtime=runtime)
 
         main_file = '%s/%s' % (lambda_cwd, handler_file)
-        if not os.path.exists(main_file):
+
+        if CHECK_HANDLER_ON_CREATION and not os.path.exists(main_file):
             # Raise an error if (1) this is not a local mount lambda, or (2) we're
             # running Lambdas locally (not in Docker), or (3) we're using remote Docker.
             # -> We do *not* want to raise an error if we're using local mount in non-remote Docker
@@ -928,6 +944,9 @@ def create_function():
     """
     arn = 'n/a'
     try:
+        if len(request.data) > FUNCTION_MAX_SIZE:
+            return error_response('Request must be smaller than %s bytes for the CreateFunction operation' %
+                FUNCTION_MAX_SIZE, 413, error_type='RequestEntityTooLargeException')
         data = json.loads(to_str(request.data))
         lambda_name = data['FunctionName']
         event_publisher.fire_event(event_publisher.EVENT_LAMBDA_CREATE_FUNC,
@@ -959,7 +978,7 @@ def create_function():
         func_details.code.pop('ZipFile', None)
         # prepare result
         result.update(format_func_details(func_details))
-        if data.get('Publish', False):
+        if data.get('Publish'):
             result['Version'] = publish_new_function_version(arn)['Version']
         return jsonify(result or {})
     except Exception as e:
@@ -1132,22 +1151,48 @@ def add_permission(function):
     data = json.loads(to_str(request.data))
     iam_client = aws_stack.connect_to_service('iam')
     sid = data.get('StatementId')
+    action = data.get('Action')
+    principal = data.get('Principal')
+    sourcearn = data.get('SourceArn')
+    arn = func_arn(function)
+
+    if not re.match(r'lambda:[*]|lambda:[a-zA-Z]+|[*]', action):
+        return error_response('1 validation error detected: Value "%s" at "action" failed to satisfy '
+                              'constraint: Member must satisfy regular expression pattern: '
+                              '(lambda:[*]|lambda:[a-zA-Z]+|[*])' % action,
+                              400, error_type='ValidationException')
+
     policy = {
         'Version': IAM_POLICY_VERSION,
         'Id': 'LambdaFuncAccess-%s' % sid,
         'Statement': [{
             'Sid': sid,
             'Effect': 'Allow',
-            # TODO: 'Principal' in policies not yet supported in upstream moto
-            # 'Principal': data.get('Principal') or {'AWS': TEST_AWS_ACCOUNT_ID},
-            'Action': data.get('Action'),
-            'Resource': func_arn(function)
+            'Action': action,
+            'Resource': arn,
         }]
     }
+
+    # Adds SourceArn only if SourceArn is present
+    if sourcearn:
+        condition = {
+            'ArnLike': {
+                'AWS:SourceArn': sourcearn
+            }
+        }
+        policy['Statement'][0]['Condition'] = condition
+
+    # Adds Principal only if Principal is present
+    if principal:
+        principal = {
+            'Service': principal
+        }
+        policy['Statement'][0]['Principal'] = principal
+
     iam_client.create_policy(PolicyName=POLICY_NAME_PATTERN % (function, sid),
                              PolicyDocument=json.dumps(policy),
                              Description='Policy for Lambda function "%s"' % function)
-    result = {'Statement': sid}
+    result = {'Statement': json.dumps(policy['Statement'][0])}
     return jsonify(result)
 
 
@@ -1196,7 +1241,6 @@ def invoke_function(function):
         arn = m.group(1)
     else:
         qualifier = request.args.get('Qualifier')
-
     data = request.get_data()
     if data:
         data = to_str(data)
@@ -1267,7 +1311,8 @@ def invoke_function(function):
         return not_found
 
     if invocation_type == 'RequestResponse':
-        result = run_lambda(asynchronous=False, func_arn=arn, event=data, context={}, version=qualifier)
+        context = {'client_context': request.headers.get('X-Amz-Client-Context')}
+        result = run_lambda(asynchronous=False, func_arn=arn, event=data, context=context, version=qualifier)
         return _create_response(result)
     elif invocation_type == 'Event':
         run_lambda(asynchronous=True, func_arn=arn, event=data, context={}, version=qualifier)
@@ -1487,11 +1532,12 @@ def untag_resource(version, arn):
     return jsonify({})
 
 
-@app.route('/2019-09-25/functions/<function>/event-invoke-config', methods=['PUT'])
+@app.route('/2019-09-25/functions/<function>/event-invoke-config', methods=['PUT', 'POST'])
 def put_function_event_invoke_config(function):
-    """ Updates the configuration for asynchronous invocation for a function
+    # TODO: resouce validation required to check if resource exists
+    """ Add/Updates the configuration for asynchronous invocation for a function
         ---
-        operationId: PutFunctionEventInvokeConfig
+        operationId: PutFunctionEventInvokeConfig | UpdateFunctionEventInvokeConfig
         parameters:
             - name: 'function'
               in: path
@@ -1503,6 +1549,9 @@ def put_function_event_invoke_config(function):
     data = json.loads(to_str(request.data))
     function_arn = func_arn(function)
     lambda_obj = ARN_TO_LAMBDA[function_arn]
+
+    if request.method == 'PUT':
+        response = lambda_obj.clear_function_event_invoke_config()
     response = lambda_obj.put_function_event_invoke_config(data)
 
     return jsonify({
@@ -1544,6 +1593,18 @@ def get_function_event_invoke_config(function):
     return jsonify(response)
 
 
+@app.route('/2019-09-25/functions/<function>/event-invoke-config', methods=['DELETE'])
+def delete_function_event_invoke_config(function):
+    try:
+        function_arn = func_arn(function)
+        lambda_obj = ARN_TO_LAMBDA[function_arn]
+    except Exception as e:
+        return error_response(str(e), 400)
+
+    lambda_obj.clear_function_event_invoke_config()
+    return Response('', status=204)
+
+
 def serve(port, quiet=True):
     from localstack.services import generic_proxy  # moved here to fix circular import errors
 
