@@ -191,7 +191,7 @@ def serve_health_endpoint(method, path, data):
 def serve_resource_graph(data):
     data = json.loads(to_str(data or '{}'))
     env = Environment.from_string(data.get('awsEnvironment'))
-    graph = dashboard_infra.get_graph(name_filter=data.get('nameFilter') or '.*', env=env)
+    graph = dashboard_infra.get_graph(name_filter=data.get('nameFilter') or '.*', env=env, region=data.get('awsRegion'))
     return graph
 
 
@@ -219,6 +219,9 @@ def get_api_from_custom_rules(method, path, data, headers):
     if path == '/' and b'QueueName=' in data_bytes:
         return 'sqs', config.PORT_SQS
 
+    if path.startswith('/2015-03-31/functions/'):
+        return 'lambda', config.PORT_LAMBDA
+
     # TODO: move S3 public URLs to a separate port/endpoint, OR check ACLs here first
     stripped = path.strip('/')
     if method in ['GET', 'HEAD'] and '/' in stripped:
@@ -229,7 +232,7 @@ def get_api_from_custom_rules(method, path, data, headers):
     if stripped and '/' not in stripped:
         if method == 'HEAD':
             # assume that this is an S3 HEAD bucket request with URL path `/<bucket>`
-            return config.PORT_S3
+            return 's3', config.PORT_S3
         if method == 'PUT':
             # assume that this is an S3 PUT bucket request with URL path `/<bucket>`
             return 's3', config.PORT_S3
