@@ -1,3 +1,5 @@
+import time
+import gzip
 import re
 import json
 import uuid
@@ -9,14 +11,20 @@ import datetime
 import xmltodict
 import collections
 import dateutil.parser
+import urllib.parse
 import six
 import botocore.config
 from pytz import timezone
+from urllib.parse import parse_qs
+from botocore.compat import urlsplit
 from botocore.client import ClientError
+from botocore.credentials import Credentials
+from localstack.utils.auth import HmacV1QueryAuth
+from botocore.awsrequest import create_request_object
 from requests.models import Response, Request
 from six.moves.urllib import parse as urlparse
 from localstack import config, constants
-from localstack.config import HOSTNAME, HOSTNAME_EXTERNAL
+from localstack.config import HOSTNAME, HOSTNAME_EXTERNAL, LOCALHOST_IP
 from localstack.utils.aws import aws_stack
 from localstack.services.s3 import multipart_content
 from localstack.utils.common import (
@@ -25,7 +33,8 @@ from localstack.utils.common import (
 from localstack.utils.analytics import event_publisher
 from localstack.utils.http_utils import uses_chunked_encoding
 from localstack.utils.persistence import PersistingProxyListener
-from localstack.utils.aws.aws_responses import requests_response
+from localstack.constants import TEST_AWS_ACCESS_KEY_ID, TEST_AWS_SECRET_ACCESS_KEY
+from localstack.utils.aws.aws_responses import requests_response, requests_error_response_xml_signature_calculation
 
 CONTENT_SHA256_HEADER = 'x-amz-content-sha256'
 STREAMING_HMAC_PAYLOAD = 'STREAMING-AWS4-HMAC-SHA256-PAYLOAD'
@@ -84,6 +93,17 @@ ALLOWED_HEADER_OVERRIDES = {
 # https://github.com/boto/botocore/blob/30206ab9e9081c80fa68e8b2cb56296b09be6337/botocore/auth.py#L47
 POLICY_EXPIRATION_FORMAT = '%Y-%m-%dT%H:%M:%SZ'
 
+# ignored_headers_lower conatins headers which don't get involved in signature calculations process
+# these headers are being sent by the localstack by default.
+IGNORED_HEADERS_LOWER = [
+    'remote-addr', 'host', 'user-agent', 'accept-encoding',
+    'accept', 'connection', 'origin',
+    'x-forwarded-for', 'x-localstack-edge', 'authorization'
+]
+
+# params are required in presigned url
+PRESIGN_QUERY_PARAMS = ['Signature', 'Expires', 'AWSAccessKeyId']
+
 
 def event_type_matches(events, action, api_method):
     """ check whether any of the event types in `events` matches the
@@ -123,7 +143,7 @@ def prefix_with_slash(s):
     return s if s[0] == '/' else '/%s' % s
 
 
-def get_event_message(event_name, bucket_name, file_name='testfile.txt', version_id=None, file_size=0):
+def get_event_message(event_name, bucket_name, file_name='testfile.txt', etag='', version_id=None, file_size=0):
     # Based on: http://docs.aws.amazon.com/AmazonS3/latest/dev/notification-content-structure.html
     bucket_name = normalize_bucket_name(bucket_name)
     return {
@@ -154,9 +174,9 @@ def get_event_message(event_name, bucket_name, file_name='testfile.txt', version
                     'arn': 'arn:aws:s3:::%s' % bucket_name
                 },
                 'object': {
-                    'key': file_name,
+                    'key': urllib.parse.quote(file_name),
                     'size': file_size,
-                    'eTag': 'd41d8cd98f00b204e9800998ecf8427e',
+                    'eTag': etag,
                     'versionId': version_id,
                     'sequencer': '0055AED6DCD90281E5'
                 }
@@ -166,9 +186,8 @@ def get_event_message(event_name, bucket_name, file_name='testfile.txt', version
 
 
 def send_notifications(method, bucket_name, object_path, version_id):
-    bucket_name = normalize_bucket_name(bucket_name)
     for bucket, notifs in S3_NOTIFICATIONS.items():
-        if bucket == bucket_name:
+        if normalize_bucket_name(bucket) == normalize_bucket_name(bucket_name):
             action = {'PUT': 'ObjectCreated', 'POST': 'ObjectCreated', 'DELETE': 'ObjectRemoved'}[method]
             # TODO: support more detailed methods, e.g., DeleteMarkerCreated
             # http://docs.aws.amazon.com/AmazonS3/latest/dev/NotificationHowTo.html
@@ -193,17 +212,20 @@ def send_notification_for_subscriber(notif, bucket_name, object_path, version_id
     key = urlparse.unquote(object_path.replace('//', '/'))[1:]
 
     s3_client = aws_stack.connect_to_service('s3')
+    object_data = {}
     try:
-        object_size = s3_client.head_object(Bucket=bucket_name, Key=key).get('ContentLength', 0)
+        object_data = s3_client.head_object(Bucket=bucket_name, Key=key)
+
     except botocore.exceptions.ClientError:
-        object_size = 0
+        pass
 
     # build event message
     message = get_event_message(
         event_name=event_name,
         bucket_name=bucket_name,
         file_name=key,
-        file_size=object_size,
+        etag=object_data.get('ETag', ''),
+        file_size=object_data.get('ContentLength', 0),
         version_id=version_id
     )
     message = json.dumps(message)
@@ -497,6 +519,17 @@ def fix_creation_date(method, path, response):
         r'\1Z</CreationDate>', to_str(response._content))
 
 
+def fix_delimiter(data, headers, response):
+    if response.status_code == 200 and response._content:
+        c, xml_prefix, delimiter = response._content, '<?xml', '<Delimiter><'
+        pattern = '[<]Delimiter[>]None[<]'
+        if isinstance(c, bytes):
+            xml_prefix, delimiter = xml_prefix.encode(), delimiter.encode()
+            pattern = pattern.encode()
+        if c.startswith(xml_prefix):
+            response._content = re.compile(pattern).sub(delimiter, c)
+
+
 def convert_to_chunked_encoding(method, path, response):
     if method != 'GET' or path != '/':
         return
@@ -507,6 +540,21 @@ def convert_to_chunked_encoding(method, path, response):
     response.headers.pop('Content-Length', None)
 
 
+def unquote(s):
+    if (s[0], s[-1]) in (('"', '"'), ("'", "'")):
+        return s[1:-1]
+    return s
+
+
+def ret304_on_etag(data, headers, response):
+    etag = response.headers.get('ETag')
+    if etag:
+        match = headers.get('If-None-Match')
+        if match and unquote(match) == unquote(etag):
+            response.status_code = 304
+            response._content = ''
+
+
 def fix_etag_for_multipart(data, headers, response):
     # Fix for https://github.com/localstack/localstack/issues/1978
     if headers.get(CONTENT_SHA256_HEADER) == STREAMING_HMAC_PAYLOAD:
@@ -796,8 +844,8 @@ def get_key_name(path, headers):
 
 
 def uses_path_addressing(headers):
-    host = headers['host']
-    return host.startswith(HOSTNAME) or host.startswith(HOSTNAME_EXTERNAL)
+    host = headers.get(constants.HEADER_LOCALSTACK_EDGE_URL, '').split('://')[-1] or headers['host']
+    return host.startswith(HOSTNAME) or host.startswith(HOSTNAME_EXTERNAL) or host.startswith(LOCALHOST_IP)
 
 
 def get_bucket_name(path, headers):
@@ -853,7 +901,7 @@ def handle_notification_request(bucket, method, data):
                     if dest in notif:
                         dest_dict = {
                             '%sConfiguration' % dest: {
-                                'Id': uuid.uuid4(),
+                                'Id': notif['Id'],
                                 dest: notif[dest],
                                 'Event': notif['Event'],
                                 'Filter': notif['Filter']
@@ -881,7 +929,7 @@ def handle_notification_request(bucket, method, data):
                     s3_filter['FilterRule'] = [s3_filter['FilterRule']]
                 # create final details dict
                 notification_details = {
-                    'Id': config.get('Id'),
+                    'Id': config.get('Id', str(uuid.uuid4())),
                     'Event': events,
                     dest: config.get(dest),
                     'Filter': event_filter
@@ -947,6 +995,17 @@ class ProxyListenerS3(PersistingProxyListener):
             return True
 
     def forward_request(self, method, path, data, headers):
+
+        # Create list of query parameteres from the url
+        parsed = urlparse.urlparse('{}{}'.format(config.get_edge_url(), path))
+        query_params = parse_qs(parsed.query)
+
+        # Detecting pre-sign url and checking signature
+        if any([p in query_params for p in PRESIGN_QUERY_PARAMS]):
+            response = authenticate_presign_url(method=method, path=path, data=data, headers=headers)
+            if response is not None:
+                return response
+
         # parse path and query params
         parsed_path = urlparse.urlparse(path)
 
@@ -1175,10 +1234,12 @@ class ProxyListenerS3(PersistingProxyListener):
                 error_doc_key = website_config.get('ErrorDocument', {}).get('Key')
 
                 if error_doc_key:
-                    error_object = s3_client.get_object(Bucket=bucket_name, Key=error_doc_key)
-                    response.status_code = 200
-                    response._content = error_object['Body'].read()
-                    response.headers['Content-Length'] = str(len(response._content))
+                    error_doc_path = '/' + bucket_name + '/' + error_doc_key
+                    if parsed.path != error_doc_path:
+                        error_object = s3_client.get_object(Bucket=bucket_name, Key=error_doc_key)
+                        response.status_code = 200
+                        response._content = error_object['Body'].read()
+                        response.headers['Content-Length'] = str(len(response._content))
             except ClientError:
                 # Pass on the 404 as usual
                 pass
@@ -1195,14 +1256,16 @@ class ProxyListenerS3(PersistingProxyListener):
             fix_metadata_key_underscores(response=response)
             fix_creation_date(method, path, response=response)
             fix_etag_for_multipart(data, headers, response)
+            ret304_on_etag(data, headers, response)
             append_aws_request_troubleshooting_headers(response)
+            fix_delimiter(data, headers, response)
 
             if method == 'PUT':
                 set_object_expiry(path, headers)
 
             # Remove body from PUT response on presigned URL
             # https://github.com/localstack/localstack/issues/1317
-            if method == 'PUT' and ('X-Amz-Security-Token=' in path or
+            if method == 'PUT' and response.status_code < 400 and ('X-Amz-Security-Token=' in path or
                     'X-Amz-Credential=' in path or 'AWSAccessKeyId=' in path):
                 response._content = ''
                 reset_content_length = True
@@ -1266,6 +1329,90 @@ class ProxyListenerS3(PersistingProxyListener):
             # convert to chunked encoding, for compatibility with certain SDKs (e.g., AWS PHP SDK)
             convert_to_chunked_encoding(method, path, response)
 
+            if headers.get('Accept-Encoding') == 'gzip':
+                if response._content is None:
+                    response._content = ''
+                response._content = gzip.compress(to_bytes(response._content))
+                response.headers['Content-Length'] = str(len(response._content))
+                response.headers['Content-Encoding'] = 'gzip'
+
+
+def authenticate_presign_url(method, path, headers, data=None):
+
+    sign_headers = []
+    url = '{}{}'.format(config.get_edge_url(), path)
+    parsed = urlparse.urlparse(url)
+    query_params = parse_qs(parsed.query)
+
+    # Checking required parameters are present in url or not
+    if not all([p in query_params for p in PRESIGN_QUERY_PARAMS]):
+        return requests_error_response_xml_signature_calculation(
+            code=403,
+            message='Query-string authentication requires the Signature, Expires and AWSAccessKeyId parameters',
+            code_string='AccessDenied'
+        )
+
+    # Fetching headers which has been sent to the requets
+    for header in headers:
+        key = header[0]
+        if key.lower() not in IGNORED_HEADERS_LOWER:
+            sign_headers.append(header)
+
+    # Request's headers are more essentials than the query parameters in the requets.
+    # Different values of header in the header of the request and in the query paramter of the requets url
+    # will fail the signature calulation. As per the AWS behaviour
+    presign_params_lower = [p.lower() for p in PRESIGN_QUERY_PARAMS]
+    if len(query_params) > 2:
+        for key in query_params:
+            if key.lower() not in presign_params_lower:
+                if key.lower() not in (header[0].lower() for header in headers):
+                    sign_headers.append((key, query_params[key][0]))
+
+    # Preparnig dictionary of request to build AWSRequest's object of the botocore
+    request_dict = {
+        'url_path': path.split('?')[0],
+        'query_string': {},
+        'method': method,
+        'headers': dict(sign_headers),
+        'body': b'',
+        'url': url.split('?')[0],
+        'context': {
+            'is_presign_request': True,
+            'use_global_endpoint': True,
+            'signing': {
+                'bucket': str(path.split('?')[0]).split('/')[1]
+            }
+        }
+    }
+    aws_request = create_request_object(request_dict)
+
+    # Calculating Signature
+    credentials = Credentials(access_key=TEST_AWS_ACCESS_KEY_ID, secret_key=TEST_AWS_SECRET_ACCESS_KEY)
+    auth = HmacV1QueryAuth(credentials=credentials, expires=query_params['Expires'][0])
+    split = urlsplit(aws_request.url)
+    string_to_sign = auth.get_string_to_sign(method=method, split=split, headers=aws_request.headers)
+    signature = auth.get_signature(string_to_sign=string_to_sign)
+
+    # Comparing the signature in url with signature we calculated
+    if query_params['Signature'][0] != signature:
+        return requests_error_response_xml_signature_calculation(
+            code=403,
+            code_string='SignatureDoesNotMatch',
+            aws_access_token=TEST_AWS_ACCESS_KEY_ID,
+            string_to_sign=string_to_sign,
+            signature=signature,
+            message='The request signature we calculated does not match the signature you provided. \
+                    Check your key and signing method.')
+
+    # Checking whether the url is expired or not
+    if int(query_params['Expires'][0]) < time.time():
+        return requests_error_response_xml_signature_calculation(
+            code=403,
+            code_string='AccessDenied',
+            message='Request has expired',
+            expires=query_params['Expires'][0]
+        )
+
 
 # instantiate listener
 UPDATE_S3 = ProxyListenerS3()
